"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BindingEngine = void 0;
const core_1 = require("./core");
const core_constants_1 = require("./core.constants");
const core_types_1 = require("./core.types");
const funcs_1 = require("./funcs");
var BindingWay;
(function (BindingWay) {
    BindingWay[BindingWay["IN_WAY"] = 0] = "IN_WAY";
    BindingWay[BindingWay["OUT_WAY"] = 1] = "OUT_WAY";
    BindingWay[BindingWay["OUTPUT_WAY"] = 2] = "OUTPUT_WAY";
    BindingWay[BindingWay["TWO_WAY"] = 3] = "TWO_WAY";
    BindingWay[BindingWay["CONSTRUCTIONAL"] = 4] = "CONSTRUCTIONAL";
    BindingWay[BindingWay["BOTH_WAY"] = 5] = "BOTH_WAY";
    BindingWay[BindingWay["NONE"] = 6] = "NONE";
})(BindingWay || (BindingWay = {}));
var BindingValue;
(function (BindingValue) {
    BindingValue[BindingValue["PROPERTY"] = 0] = "PROPERTY";
    BindingValue[BindingValue["STATIC"] = 1] = "STATIC";
    BindingValue[BindingValue["FUNCTION"] = 2] = "FUNCTION";
})(BindingValue || (BindingValue = {}));
const FUNCTION_REGEXP = new RegExp(/\(.*\)$/);
const FUNCTION_ARGS_REGEXP = new RegExp(/\([0-9,A-a,Z-z,\., \,]+\)$/);
class BindingEngine {
    constructor(controllerInstance) {
        this.controllerInstance = controllerInstance;
    }
    getControllerBindedToNodes(controllerInstance, parentNode) {
        controllerInstance !== null && controllerInstance !== void 0 ? controllerInstance : (controllerInstance = this.controllerInstance);
        const target = parentNode.cloneNode(false);
        target.guid = parentNode.guid;
        parentNode.childNodes.forEach((childNodeOld) => {
            const childNodeNew = this.processNode(controllerInstance, childNodeOld);
            target.append(...childNodeNew);
        });
        return [controllerInstance, target];
    }
    processNode(controllerInstance, node, customVariable = null) {
        var _a, _b, _c, _d, _e, _f, _g;
        const el = node;
        const controller = this.importControllerByTag(controllerInstance, el.tagName);
        (_a = el.guid) !== null && _a !== void 0 ? _a : (el.guid = (0, funcs_1.guid)());
        let pack = [];
        if (controller == null) {
            // console.log("now scanning", node, node.nodeType);
            if (node.tagName == "PRE") {
                pack.push(node);
            }
            else {
                switch (node.nodeType) {
                    case node.ELEMENT_NODE:
                        const attribute_names_all = el.getAttributeNames();
                        const attribute_names = attribute_names_all.filter(el => !el.startsWith(core_constants_1.CONSTRUCTIONAL));
                        const constructional_names = attribute_names_all.filter(el => el.startsWith(core_constants_1.CONSTRUCTIONAL));
                        if (constructional_names.length > 0) {
                            for (const constructional_name of constructional_names) {
                                return this.processNodeConstructionally(controllerInstance, el, constructional_name);
                            }
                        }
                        else {
                            const template_element = node.cloneNode(false);
                            template_element.guid = node.guid;
                            for (const attribute_name of attribute_names) {
                                const attribute_value = (_c = (_b = template_element.getAttribute(attribute_name)) === null || _b === void 0 ? void 0 : _b.trim()) !== null && _c !== void 0 ? _c : '';
                                const attribute_name_specless = this.withoutSpec(attribute_name);
                                if (attribute_name_specless != attribute_name) {
                                    template_element.removeAttribute(attribute_name);
                                }
                                switch (this.checkBindingWay(attribute_name, attribute_value)) {
                                    case BindingWay.IN_WAY:
                                        this.bindInWay(attribute_name_specless, attribute_value, template_element, controllerInstance, customVariable);
                                        break;
                                    case BindingWay.OUT_WAY:
                                        this.bindOutWay(attribute_name_specless, attribute_value, template_element, controllerInstance, customVariable);
                                        break;
                                    case BindingWay.BOTH_WAY:
                                        this.bindBothWay(attribute_name_specless, attribute_value, template_element, controllerInstance, customVariable);
                                        break;
                                    case BindingWay.NONE:
                                        break;
                                    default:
                                        throw new Error(`Can't identify binding type '${attribute_name}' '${attribute_value}'`);
                                }
                            }
                            node.childNodes.forEach((childNodeOld) => {
                                const childNodeNew = this.processNode(controllerInstance, childNodeOld, customVariable);
                                template_element.append(...childNodeNew);
                            });
                            pack = [template_element];
                        }
                        break;
                    case node.TEXT_NODE:
                        // console.log(node.parentElement);
                        const contents = (_e = (_d = node.textContent) === null || _d === void 0 ? void 0 : _d.trim().split(' ')) !== null && _e !== void 0 ? _e : [];
                        // const parent = node.parentNode!;
                        for (let textToken of contents) {
                            const newNode = document.createTextNode(textToken);
                            newNode.guid = (0, funcs_1.guid)();
                            pack.push(newNode, document.createTextNode(' '));
                            switch (this.checkBindingWay('', textToken)) {
                                case BindingWay.OUTPUT_WAY:
                                    this.bindTextOutPutWay(this.withoutSpec(textToken), newNode, controllerInstance, customVariable);
                                default:
                                    // parent.insertBefore(newNode, node)
                                    continue;
                            }
                        }
                    default:
                    // console.error(node);
                    // throw new Error(`Can't determine node type ${node.nodeType}`);
                }
            }
        }
        else {
            const controllerManager = new core_1.uwuControllerManager();
            let [controllerInstanceInner, el] = controllerManager.processPageByController(controller, node);
            if ((0, core_1.declarationCheck)(controller.prototype.type, core_types_1.uwuDeclaration.Controller)) {
                const bindingEngine = new BindingEngine(controllerInstanceInner);
                [controllerInstanceInner, el] = bindingEngine.getControllerBindedToNodes(controllerInstanceInner, el);
                for (const attribute_name of el.getAttributeNames()) {
                    const attribute_value = (_g = (_f = el.getAttribute(attribute_name)) === null || _f === void 0 ? void 0 : _f.trim()) !== null && _g !== void 0 ? _g : '';
                    const attribute_name_specless = this.withoutSpec(attribute_name);
                    switch (this.checkBindingWay(attribute_name, attribute_value)) {
                        case BindingWay.IN_WAY:
                            this.bindInWayInner(attribute_name_specless, attribute_value, el, controllerInstanceInner);
                            break;
                        case BindingWay.OUT_WAY:
                            this.bindOutWay(attribute_name_specless, attribute_value, el, controllerInstance);
                            break;
                        case BindingWay.NONE:
                            break;
                        default:
                            throw new Error(`Can't identify binding type for inner controller '${attribute_name}' '${attribute_value}'`);
                    }
                }
            }
            pack = [el];
        }
        return pack;
    }
    processNodeConstructionally(controllerInstance, el, attribute_name) {
        const attribute_name_specless = this.withoutSpec(attribute_name);
        const attribute_value = el.getAttribute(attribute_name);
        const node_template = el.cloneNode(true);
        node_template.removeAttribute(attribute_name);
        const template_comment = document.createComment(el.guid);
        if (attribute_value == null) {
            throw new Error(`Empty attribute value faced ${attribute_name} in ${el}`);
        }
        switch (attribute_name_specless) {
            case `${core_constants_1.KEYWORD}for`:
                const declaration = attribute_value.trim().split(' ');
                if (declaration[2] != 'of' || declaration[0] != 'let' || declaration.length != 4) {
                    throw new Error("Bad uFor declaration, must be 'let element of this.array'");
                }
                const [l, elementName, o, targetArray] = declaration;
                const elementsArray = [];
                let array = this.scopeEval(controllerInstance, targetArray);
                if (!Array.isArray(array)) {
                    throw new Error(`${targetArray} is not type of Array`);
                }
                const action = (newArray) => {
                    var _a;
                    for (const node of elementsArray) {
                        (_a = template_comment.parentElement) === null || _a === void 0 ? void 0 : _a.removeChild(node);
                    }
                    while (elementsArray.length) {
                        elementsArray.pop();
                    }
                    for (const newElement of newArray) {
                        const htmlElement = node_template.cloneNode(true);
                        htmlElement.guid = (0, funcs_1.guid)();
                        const innerElements = this.processNode(controllerInstance, htmlElement, {
                            element: newElement,
                            elementName: elementName,
                            elementAction: action,
                            elementArray: array,
                        });
                        template_comment.before(...innerElements);
                        elementsArray.push(...innerElements);
                    }
                };
                (0, funcs_1.observeArray)(array, action);
                action(this.scopeEval(controllerInstance, targetArray));
                return [template_comment, ...elementsArray];
            default:
                throw new Error(`Can't find attribute ${attribute_name}`);
        }
    }
    importControllerByTag(controllerInstance, tagName) {
        var _a;
        const module = controllerInstance.parent;
        return (_a = module.prototype.exports.find((el) => { var _a, _b; return tagName != undefined && ((_b = (_a = el.prototype) === null || _a === void 0 ? void 0 : _a.selector.toString().toLowerCase()) !== null && _b !== void 0 ? _b : "") == tagName.toLowerCase(); })) !== null && _a !== void 0 ? _a : null;
    }
    borderedWith(target, left, right) {
        return target.startsWith(left) && target.endsWith(right);
    }
    withoutSpec(target) {
        for (const spec of core_constants_1.SPECS) {
            target = target.replace(spec, '');
        }
        return target;
    }
    scopeEval(scope, script) {
        return Function('"use strict";return (' + script + ')').bind(scope)();
    }
    checkBindingWay(attribute_name, attribute_value) {
        attribute_name = attribute_name.trim();
        attribute_value = attribute_value.trim();
        if (attribute_value != null) {
            if (this.borderedWith(attribute_name, core_constants_1.BOTH_WAY_BIND_OPEN, core_constants_1.BOTH_WAY_BIND_CLOSE)) {
                return BindingWay.BOTH_WAY;
            }
            else if (this.borderedWith(attribute_value, core_constants_1.CONSTRUCTIONAL, '')) {
                return BindingWay.CONSTRUCTIONAL;
            }
            else if (this.borderedWith(attribute_name, core_constants_1.IN_WAY_BIND_OPEN, core_constants_1.IN_WAY_BIND_CLOSE)) {
                return BindingWay.IN_WAY;
            }
            else if (this.borderedWith(attribute_name, core_constants_1.OUT_WAY_BIND_OPEN, core_constants_1.OUT_WAY_BIND_CLOSE)) {
                return BindingWay.OUT_WAY;
            }
            else if (this.borderedWith(attribute_value, core_constants_1.OUTPUT_WAY_BIND_OPEN, core_constants_1.OUTPUT_WAY_BIND_CLOSE)) {
                return BindingWay.OUTPUT_WAY;
            }
        }
        return BindingWay.NONE;
    }
    bindBothWay(attribute_name_specless, attribute_value, el, controllerInstance = this.controllerInstance, customVariable = null) {
        switch (attribute_name_specless) {
            case `${core_constants_1.KEYWORD}model`:
                switch (el.tagName) {
                    case 'INPUT':
                        const actionIn = this.getActionByAttribute(el, 'value');
                        const actionOut = (ev) => this.dispatchBindingEvent(controllerInstance, attribute_value, ev.target.value);
                        this.addBindingEventToAttribute(attribute_value, actionIn, el, controllerInstance, customVariable);
                        this.bindOutWay('keyup', attribute_value, el, controllerInstance, customVariable, actionOut);
                        break;
                    default:
                        throw new Error(`Can't find model binding way for element ${el}`);
                }
                break;
            default:
                throw new Error(`Can't find binding way for ${attribute_name_specless}`);
        }
    }
    bindInWay(attribute_name_specless, attribute_value, el, controllerInstance = this.controllerInstance, customVariable = null) {
        const action = this.getActionByAttribute(el, attribute_name_specless);
        this.addBindingEventToAttribute(attribute_value, action, el, controllerInstance, customVariable);
    }
    bindInWayInner(attribute_name_specless, attribute_value, el, controllerInstance = this.controllerInstance) {
        const attribute_name_specless_this = `this.${attribute_name_specless}`;
        const action = (newValue) => this.dispatchBindingEvent(controllerInstance, attribute_name_specless_this, newValue);
        this.addBindingEventToAttribute(attribute_value, action, el, this.controllerInstance);
    }
    bindOutWay(attribute_name_specless, attribute_value, el, controllerInstance = this.controllerInstance, customVariable = null, action = null) {
        let [cleared_attribute_value, args] = this.getFunctionClearedAttributeValue(attribute_value);
        el.addEventListener(attribute_name_specless, (ev) => {
            const propToRun = this.scopeEval(controllerInstance, cleared_attribute_value);
            if (typeof propToRun == 'function') {
                args = this.scopeEval(controllerInstance, `[${args}]`);
                propToRun.call(controllerInstance, ev, ...args);
            }
            else if (action != null) {
                action(ev, ...args);
            }
            else {
                this.scopeEval(controllerInstance, cleared_attribute_value);
            }
        });
        return el;
    }
    bindTextOutPutWay(attribute_value, el, controllerInstance = this.controllerInstance, customVariable = null) {
        const action = (newValue) => el.textContent = newValue;
        this.addBindingEventToAttribute(attribute_value, action, el, controllerInstance, customVariable);
        return el;
    }
    addBindingEventToAttribute(attribute_value, action, elementToBind, controllerInstance = this.controllerInstance, customVariable = null) {
        var _a, _b, _c, _d, _e;
        // Clearing calls from value
        let [cleared_attribute_value, _] = this.getFunctionClearedAttributeValue(attribute_value);
        // Getting parts for replacing to ''
        let tokens = cleared_attribute_value.split('.');
        const fullTokens = [];
        tokens.reduce((prev, next) => {
            const joined = [prev, next].join('.');
            fullTokens.push(joined);
            return joined;
        });
        if (customVariable != null && cleared_attribute_value.startsWith(customVariable.elementName)) {
            const scopedCustomVariableName = cleared_attribute_value.replace(customVariable.elementName, 'this');
            let currentValue = this.scopeEval(customVariable.element, scopedCustomVariableName);
            action(currentValue);
        }
        else if (fullTokens.length > 0) {
            // Go through every part and add it to bindPack
            // console.log("Starting binding", cleared_attribute_value, "to", elementToBind);
            for (const token of fullTokens) {
                const shouldRunAction = cleared_attribute_value == token;
                const parentProperty = this.scopeEval(controllerInstance, token.slice(0, token.lastIndexOf('.')));
                const splittedTokens = token.split('.');
                const propertyToBind = splittedTokens[splittedTokens.length - 1];
                let currentValue = this.scopeEval(controllerInstance, token);
                if (elementToBind.guid == undefined) {
                    throw new Error("Trying to bind element with empty guid");
                }
                if (token in controllerInstance.bindPack && (((_a = elementToBind.guid) !== null && _a !== void 0 ? _a : "") in controllerInstance.bindPack[token].actions == false)) {
                    // console.log("value was in bindPack and guid was not set in actions for this guid", token, elementToBind);
                    if (shouldRunAction) {
                        controllerInstance.bindPack[token].actions[elementToBind.guid] = action;
                    }
                }
                else {
                    // console.log("setting binding at first", token, elementToBind);
                    Object.defineProperty(parentProperty, propertyToBind, {
                        set: (newValue) => {
                            currentValue = newValue;
                            this.dispatchBindingEvent(controllerInstance, token);
                        },
                        get: () => currentValue,
                    });
                    controllerInstance.bindPack[token] = {
                        descriptor: Object.getOwnPropertyDescriptor(parentProperty, propertyToBind),
                        actions: shouldRunAction ? { [(_b = elementToBind.guid) !== null && _b !== void 0 ? _b : ""]: action } : {}
                    };
                }
                if (shouldRunAction) {
                    parentProperty[propertyToBind] = currentValue;
                }
            }
        }
        else {
            // just add notifiers
            const newTokens = cleared_attribute_value.split(/\W+/).filter(e => e.startsWith('this'));
            const shouldRunAction = true;
            for (const token of newTokens) {
                if (token in controllerInstance.bindPack && (((_c = elementToBind.guid) !== null && _c !== void 0 ? _c : "") in controllerInstance.bindPack[token].actions == false)) {
                    if (shouldRunAction) {
                        controllerInstance.bindPack[token].actions[(_d = elementToBind.guid) !== null && _d !== void 0 ? _d : ""] = action;
                    }
                }
                else {
                    controllerInstance.bindPack[token] = {
                        actions: shouldRunAction ? { [(_e = elementToBind.guid) !== null && _e !== void 0 ? _e : ""]: action } : {}
                    };
                }
            }
            action(this.scopeEval(controllerInstance, attribute_value));
        }
    }
    dispatchBindingEvent(controllerInstance, token, forceValue = null) {
        const bindingsToInform = Object.keys(controllerInstance.bindPack).filter(el => el.startsWith(token) /*&& el != token */);
        for (const bindingKey of bindingsToInform) {
            const binding = controllerInstance.bindPack[bindingKey];
            for (const guid in binding.actions) {
                if (binding.descriptor != undefined && binding.descriptor.get != undefined) {
                    binding.actions[guid](forceValue !== null && forceValue !== void 0 ? forceValue : binding.descriptor.get());
                }
            }
        }
    }
    getFunctionClearedAttributeValue(attribute_value) {
        var _a, _b;
        let cleared_attribute_value = attribute_value;
        (_a = FUNCTION_REGEXP.exec(attribute_value)) === null || _a === void 0 ? void 0 : _a.forEach((part) => {
            cleared_attribute_value = attribute_value.replace(part, '');
        });
        const argsArray = (_b = FUNCTION_ARGS_REGEXP.exec(attribute_value)) !== null && _b !== void 0 ? _b : [];
        const args = argsArray.length > 0 ? this.withoutSpec(argsArray[0]) : "";
        return [cleared_attribute_value, args];
    }
    getActionByAttribute(el, attribute) {
        switch (attribute.toLowerCase()) {
            case 'value':
                if (el instanceof HTMLInputElement) {
                    return (newValue) => el.value = newValue;
                }
            case 'innerhtml':
                return (newValue) => el.innerHTML = newValue;
            case `${core_constants_1.KEYWORD}class`:
                return (newValue) => {
                    Object.keys(newValue).filter(c => newValue[c]).forEach(styleClass => el.classList.add(styleClass));
                    Object.keys(newValue).filter(c => !newValue[c]).forEach(styleClass => el.classList.remove(styleClass));
                };
            case `${core_constants_1.KEYWORD}style`:
                return (newValue) => {
                    Object.keys(newValue).forEach(style => style in el.style ? el.style.setProperty(style, newValue[style]) : false);
                };
            default:
                return (newValue) => el.setAttribute(attribute, newValue);
        }
    }
}
exports.BindingEngine = BindingEngine;
//# sourceMappingURL=core.engine.js.map